import 'package:flutter/material.dart';
import 'package:osam/domain/state/base_state.dart';
import 'package:osam/domain/store/store.dart';
import 'package:osam/presentation/presenter.dart';
import 'package:provider/provider.dart';
import 'package:provider/single_child_widget.dart';

typedef LifeCycleHandler = void Function(AppLifecycleState state);

class StoreProvider<ST extends BaseState<ST>> extends StatelessWidget {
  final Store<ST> store;
  final Widget child;

  const StoreProvider({Key key, @required this.store, @required this.child}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return _InternalStoreProvider(
        store: store, child: store.isPersistEnabled ? _LifecycleWrapper<ST>(child: child) : child);
  }
}

class _InternalStoreProvider extends InheritedWidget {
  const _InternalStoreProvider({
    Key key,
    @required this.child,
    @required this.store,
  }) : super(key: key, child: child);

  final Store store;
  @override
  final Widget child;

  static _InternalStoreProvider of(BuildContext context) =>
      context.dependOnInheritedWidgetOfExactType(aspect: _InternalStoreProvider);

  @override
  bool updateShouldNotify(_InternalStoreProvider oldWidget) => store != oldWidget.store;
}

class _LifecycleWrapper<ST extends BaseState<ST>> extends StatefulWidget {
  final Widget child;

  const _LifecycleWrapper({Key key, this.child}) : super(key: key);

  @override
  _LifecycleWrapperState createState() => _LifecycleWrapperState<ST>();
}

class _LifecycleWrapperState<ST extends BaseState<ST>> extends State<_LifecycleWrapper>
    with WidgetsBindingObserver {
  @override
  void initState() {
    WidgetsBinding.instance.addObserver(this);
    super.initState();
  }

  @override
  Widget build(BuildContext context) => widget.child;

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    if (state == AppLifecycleState.paused) {
      final store = _InternalStoreProvider.of(context).store;
      store.saveState();
    }
  }
}

mixin _PresenterProviderSingleChildWidget on SingleChildWidget {}

class PresenterProvider<P extends Presenter<BaseState>> extends SingleChildStatelessWidget
    with _PresenterProviderSingleChildWidget {
  final P presenter;
  final Widget child;

  const PresenterProvider({Key key, @required this.presenter, this.child})
      : super(key: key, child: child);

  static P of<P extends Presenter<BaseState>>(BuildContext context) =>
      Provider.of<P>(context, listen: false);

  @override
  Widget buildWithChild(BuildContext context, Widget child) {
    final store = _InternalStoreProvider.of(context).store;
    assert(store != null);
    if (store == null)
      debugPrint('insure you provided store by StoreProvider in the start of your widget tree');
    return Provider(
      key: ValueKey('PresenterProvider: %${key.toString()}'),
      create: (_) => presenter
        // ignore: invalid_use_of_protected_member
        ..store = store
        ..init(),
      child: child,
      dispose: (ctx, presenter) => presenter.dispose(),
    );
  }
}

class MultiPresenterProvider extends StatelessWidget {
  final List<SingleChildWidget> providers;
  final Widget child;

  const MultiPresenterProvider({Key key, @required this.providers, @required this.child})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    final store = _InternalStoreProvider.of(context).store;
    assert(store != null);
    if (store == null)
      debugPrint('insure you provided store by StoreProvider in the start of your widget tree');
    return MultiProvider(
      providers: providers,
      key: ValueKey('PresenterMultiProvider: %${key.toString()}'),
      child: child,
    );
  }
}

extension PresenterProviderExtension on BuildContext {
  P presenter<P extends Presenter<BaseState>>() => PresenterProvider.of<P>(this);
}
